//----------------------------------------------------------------------
/// 
/// Boosted Higgs tagging
///
/// Template Overlap example program, illustration of carrying out boosted
/// Higgs subjet identification analysis
///
/// It illustrates two kinds of functionality: 
///
///  - using a boosted higgs tagger, that combines both 2-body and 3-body templates
///  - following information from jet shapes, such as planar flow
///
/// This kind of functionality was used in arXiv:1212.2977 and arXiv:1112.1957
///  for boosted Higgs searches,
/// and related functionality was used in arXiv:1211.2202 (ATLAS) in searching for boosted tops
///
/// run it with    : ./example5 WH8TeV.dat
///
/// Source code: example5.cc
//----------------------------------------------------------------------

#include "matching.hh"
#include "TemplateTagger.hh"
#include "plot2.hh"


#include "fastjet/FunctionOfPseudoJet.hh"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <time.h> // For timing
#include <iomanip> 
#include <stdio.h> // For pipe

#include "Settings.hh"



using namespace std;

using namespace TemplateOverlap;

unsigned int WhichPtBin (double PtJet, double PtMin, double PtBinWidth, int nptbins)
{
if (PtJet < PtMin) return 0;
int i =  int(floor  ((PtJet - PtMin)/PtBinWidth));
if (i > nptbins) i = nptbins;
if (i<0) return 0;
return i;
}



///  Helper function for printing jets
void PrintJets(const vector <fastjet::PseudoJet>& jets) {
   printf("%5s %15s %15s %15s %15s %15s\n","jet #", "rapidity", "phi", "pt","m","e"); // label the columns
   for (unsigned int i = 0; i < jets.size(); i++) {
      printf("%5u %15.8f %15.8f %15.8f %15.8f %15.8f\n",i, jets[i].rap(),jets[i].phi(),jets[i].perp(),jets[i].m(),jets[i].e());
   }
}


//----------------------------------------------------------------------
// forward declaration for printing out info about a jet
//----------------------------------------------------------------------
ostream & operator<<(ostream &, const fastjet::PseudoJet &);


//----------------------------------------------------------------------
// class for reading in a jet
//----------------------------------------------------------------------

bool ReadEvent (ifstream& is, vector<fastjet::PseudoJet> & input_particles, vector<fastjet::PseudoJet> & partons, fastjet::PseudoJet & lepton, fastjet::PseudoJet & met)
{

  string line;
  double pup1, pup2, pup3, pup4;
  int nhep, i, isthep,idhep,evtnumber;
  
  input_particles.resize(0);
  partons.resize(0);
  std::vector<fastjet::PseudoJet> leptons;
  
  fastjet::PseudoJet met1;
    
  if (!getline(is,line)) return false;	
  istringstream getpro(line);
  getpro  >> evtnumber >> nhep;

  for (int j=0; j<nhep;++j){
    if (!getline(is,line)) return false;	
    istringstream getall(line);
    getall  >> i >> isthep >>idhep>> pup1 >> pup2 >> pup3 >> pup4;
	
    fastjet::PseudoJet pJet = fastjet::PseudoJet(pup1,pup2,pup3,pup4);
    pJet.set_user_index(idhep);      

    if ( isthep == 21 && j > 6 ) {partons.push_back(pJet);continue;}

    if (abs(idhep) == 12 || abs(idhep) == 14 || abs(idhep) == 16) {met1 += pJet; continue;}
    if (abs(idhep) == 11 || abs(idhep) == 13) {leptons.push_back(pJet); continue;}
    input_particles.push_back(pJet);
  } 
  std::vector<fastjet::PseudoJet> sorted_leptons = sorted_by_pt(leptons);
  if (leptons.size()>0)
  {
    for (unsigned int i = 1; i < sorted_leptons.size(); ++i)
	input_particles.push_back(sorted_leptons[i]);
    lepton = sorted_leptons[0];
  }
  met = met1;
  return evtnumber;
}

void splitPartons(const std::vector<fastjet::PseudoJet> & partons, std::vector<fastjet::PseudoJet> & top1, std::vector<fastjet::PseudoJet> & top2)
{
  top1.resize(0);
  top2.resize(0);
  for (unsigned int i=0;i<partons.size();++i)
  {
    int id = partons[i].user_index();
    if (id==5|| id==2 || id==4 || id==-1 ||id==-3|| id==-11 || id==-13) top1.push_back(partons[i]);
    if (id==-5|| id==-2 || id==-4|| id==1 ||id==3 || id==11 || id==13) top2.push_back(partons[i]);

  }
  //top1 = Hadronic; top2= Leptonic
  bool foundLepton = false;
  for (unsigned int i = 0; i < top1.size(); ++i )
    if (abs(top1[i].user_index()) ==11 || abs(top1[i].user_index()) ==13) {foundLepton = true; break;}
  
  if (foundLepton) 
  {
    std::vector<fastjet::PseudoJet> temp = top1;
    top1 = top2;
    top2 = temp;
  }
  
}

//---------------------------------------------------------
/////////////////Add pileup
//---------------------------------------------------------

/// This adds a fix number of pp interactions, useful when making plots
void addPileup(std::vector<fastjet::PseudoJet> & particles, std::ifstream & pileupFile, const int nPileup)
{
  std::vector <fastjet::PseudoJet> pileups, partons;
  fastjet::PseudoJet lepton, met;
  for (int iPileup = 0; iPileup < nPileup; ++iPileup) 
  {
    if(!ReadEvent(pileupFile,pileups, partons, lepton, met)) 
    {
      pileupFile.clear() ;
      pileupFile.seekg(0, ios::beg) ;
      ReadEvent(pileupFile,pileups, partons, lepton, met);
      
    }
    for (unsigned int i = 0; i < pileups.size(); ++i)
      particles.push_back(pileups[i]);
  }
}

int main(int argc, char* argv[]) {
  
  if (argc !=2) {
    cout << "Usage:  "<< argv[0] << " <input.dat>" << endl;
    return 1;
  }
  ///-----------------------------------------------------------------
  /// Read parameters from param.dat file
  /// Many of these are not useful in this example
  ///----------------------------------------------------------------
  Settings settings;
  settings.readFile("param.dat") ;
 
  double coneRadius =  settings.parm("coneRadius"), 
      coneRadius2 =  settings.parm("coneRadius2"), 
      subConeRadius2 =  settings.parm("subConeRadius2"), 
      subConeRadius3 = settings.parm("subConeRadius3"),
      subConeComparison2 = settings.parm("subConeComparison2"),
      sigma = settings.parm("sigma");
  bool VariableCone = settings.mode("variablecone");
  double pTbinWidth = settings.parm ("ptbinwidth");
	
  double nPileupAvg = settings.parm("nPileupAvg"), minPt_clus = settings.parm("minPtclus");
  double minPtW = settings.parm("minPtW"), minMET = settings.parm("minMET");	
  double minPt = settings.parm("minPt"), 
      maxPt =settings.parm("maxPt"), 
      minMass = settings.parm("minmass"), 
      maxMass = settings.parm("maxmass"); 
  int njetoutside = settings.mode("njetoutside");
    

  double ptjetmin = settings.parm("ptjetmin");
  int nptbins = settings.mode("nptbins");

  string fileName2[] = {settings.word("file2b1"),
      settings.word("file2b2"),
      settings.word("file2b3"),
      settings.word("file2b4"),
      settings.word("file2b5"),
      settings.word("file2b6"),            
      settings.word("file2b7")};            		    
  string fileName3[] = {settings.word("file3b1"),
      settings.word("file3b2"),
      settings.word("file3b3"),
      settings.word("file3b4"),
      settings.word("file3b5"),
      settings.word("file3b6"),
      settings.word("file3b7")};

  ///--------------------------------------------------------------------------------

  
   /// Set Run Parameters 
  double ovcut = 0.; /// For TemplateTagger
  int nPileup = 0; /// For pileup overlay
 
 /// Create best matched template containers
  std::vector<fastjet::PseudoJet> maxTempl;

 /// Create event containers
  std::vector<fastjet::PseudoJet> input_particles, partons;
  fastjet::PseudoJet lepton, met;
  std::vector<fastjet::PseudoJet> myHiggs, myW;

 /// Where to read the events from
  ifstream file(argv[1]);
  ifstream pileupFile("pileup.dat");
 
 /// create 7 taggers (2-body) and 7 filters (3-body)
 ///----------------------------------------------------------
  std::vector<fastjet::Transformer *> HadTopTaggers;
  std::vector<fastjet::Transformer *> LepTopTaggers;
  
  TemplateOverlap::TemplateDefinition HadTopDef(subConeRadius3, sigma, VARIABLE, CONE, ovcut);
  TemplateOverlap::TemplateDefinition LepTopDef(subConeRadius3, sigma, VARIABLE, TOP, ovcut);
  
  for( int i=0; i<nptbins;++i)
  {
    HadTopTaggers.push_back(new fastjet::TemplateTagger(fileName2[i], HadTopDef, fastjet::SelectorMassRange(minMass,maxMass)));
    LepTopTaggers.push_back(new fastjet::TemplateTagger(fileName3[i], LepTopDef, fastjet::SelectorMassRange(minMass,maxMass)));
  }
 
  /// read in data in format i isthep idhep px py pz E 

  while(ReadEvent(file, input_particles, partons, lepton, met)){
   clock_t start = clock(); // start clock

   /// Jet definitions 
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::antikt_algorithm, coneRadius,
                                      fastjet::E_scheme, fastjet::Best);
 
  fastjet::JetDefinition         *jetDef2 = NULL;
  jetDef2 = new fastjet::JetDefinition(fastjet::antikt_algorithm, coneRadius2,
                                      fastjet::E_scheme, fastjet::Best);
  /// Overlay pileup
  if (nPileup > 0) addPileup(input_particles, pileupFile, nPileup);
  
   /// Find antikt jets
  vector <fastjet::PseudoJet> fatjets;
  fastjet::ClusterSequence clustSeq(input_particles, *jetDef);    
  fatjets = sorted_by_pt(clustSeq.inclusive_jets(25.));
  
   /// Find antikt 0.4 subjets
    vector <fastjet::PseudoJet> pJets;
    fastjet::ClusterSequence clustSeq2(input_particles, *jetDef2);
    pJets = sorted_by_pt(clustSeq2.inclusive_jets(20.));
    
    int jb=-1;
    //Count jets outside fat jet
    for (unsigned int j=0;j<pJets.size();++j )
        if (pJets[j].plain_distance(fatjets[0]) > (coneRadius+coneRadius2) * (coneRadius+coneRadius2)
      && pJets[j].perp() > ptjetmin && pJets[j].plain_distance(lepton) < coneRadius*coneRadius) {jb = j;break;}
      
      if (jb==-1) {std::cout << "No leptonic top found!" <<std::endl; continue;}

  
  /// Choose a tagger (and a filter) based on jet pT
  int iTemp = WhichPtBin(fatjets[0].perp(), minPtW, pTbinWidth, nptbins-1); 
  const fastjet::Transformer & fhad = *HadTopTaggers[iTemp];
  const fastjet::Transformer & flep = *LepTopTaggers[iTemp]; 
  std::cout << "Using Template Library "<< iTemp << std::endl;
  
  /// Pick up the hadronic and leptonic top jet candidate
  fastjet::PseudoJet HadTopJet = fatjets[0];
  
  fastjet::PseudoJet LepTopJet = join(pJets[jb], lepton, met);
  system("clear");
  

  
  /// Print jet information 
 // cout << "Run: " << jetDef->description() << endl << endl;
  cout << "Hadronic Top: " << HadTopJet << endl;
  cout << "Leptonic Top: "<< LepTopJet << endl;
  
  /// Now tag the leading jet using template tagger
  /// For the Higgs, we combine 2-body and 3-body templates
  fastjet::PseudoJet tagged1 = fhad(HadTopJet);
  fastjet::PseudoJet tagged2 = flep(LepTopJet);
  
  /// Create display file

  std::ofstream fout("sample_event_display.dat",std::ifstream::out );   
  std::ofstream fout2("sample_event_out.dat",std::ifstream::out );  
  std::ofstream fout3("sample_event_parton.dat",std::ifstream::out );  
  
  std::ofstream gout("sample_event_display2.dat",std::ifstream::out );   
  std::ofstream gout2("sample_event_out.dat2",std::ifstream::out );  
  std::ofstream gout3("sample_event_parton2.dat",std::ifstream::out );  
  
  std::vector<fastjet::PseudoJet> top1, top2;
  splitPartons(partons, top1, top2);

  
  output_vec_pseudojet(fout2,input_particles);
  
  output_vec_pseudojet(gout2,input_particles);

  
  output_vec_pseudojet(fout3, top1);
  
  output_vec_pseudojet(gout3, top2);

 // fout3<<std::endl;
 // output_vec_pseudojet(fout3, top2);

  
  /// Show me what you got
  if (tagged1 == 0)
  {
    std::cout << "No 3-body substructure found in Hadronic Top" << std::endl;
  }  
  else
  {

    
   /// We extract substructure information from filtered jet  
            
  const fastjet::TemplateTagger::StructureType & templ3b = tagged1.structure_of<fastjet::TemplateTagger>(); 
  
  fastjet::PseudoJet parent1 = tagged1.pieces()[0];
  fastjet::PseudoJet parent2 = tagged1.pieces()[1];
  fastjet::PseudoJet parent3 = tagged1.pieces()[2];

  std::cout << "Found suitable filtered subjets: " << std::endl;
  std::cout << " " << parent1 << std::endl;
  std::cout << " " << parent2 << std::endl;
  std::cout << " " << parent3 << std::endl;

  std::cout << "Total = " << std::endl;
  std::cout << " " << tagged1 << std::endl;
  std::cout << "(Ov3 = " << templ3b.ov() <<")" << std::endl << std::endl;
  std::cout << " The best-matched (3-body) template is: "<< std::endl;     
  PrintJets(templ3b.maxTempl());
  

  // Print additional information
  // Below we show how to compute planar flow 
  // directly from the jet constituents
  // or, alternatively, from the template partons
  
  // Extract the peak 3-body template
  std::vector<fastjet::PseudoJet> my3bTemplate = templ3b.maxTempl();
  
  
  // Calculate the template radius from template parton pT
  
  std::vector<double> radii = templateRadii(my3bTemplate, subConeRadius3);
 // double r1 = eShape(my3bTemplate[0].perp()) - eShape(100.) + subConeRadius3;
//  double r2 = eShape(my3bTemplate[1].perp()) - eShape(100.) + subConeRadius3;
//  double r3 = eShape(my3bTemplate[2].perp()) - eShape(100.) + subConeRadius3;
 double r1 = radii[0], r2 = radii[1], r3 = radii[2];
  // Define two planar flow functions
  // This definition of planar flow smears the energy according to the template subcones
  fastjet::planarflow pf(r1,r2,r3);
  // This computes the actual planar flow
  fastjet::planarflow pf2(0,0,0);
  
  
    std::stringstream ss; 
  ss << "(Ov3 = "<<tagged1.structure_of<fastjet::TemplateTagger>().ov() <<")";
  
     // for event display
  set_header_for_display(fout, tagged1.eta(), tagged1.phi_std(), coneRadius, ss.str() );
 //   set_header_for_display(fout,0, 0, 6.28, ss.str() );

  std::vector<fastjet::PseudoJet> constituents = HadTopJet.constituents();
  output_hadron_to_display(fout,constituents);
  
  circle(fout,tagged1,coneRadius,6, "title \"hTop\"");
  circle(fout,my3bTemplate[0],r1,3, "title \"3b-templ\"");
  circle(fout,my3bTemplate[1],r2,3, "notitle");
  circle(fout,my3bTemplate[2],r3,3, "notitle");

  
 // std::cout << "3-body template radii " << r1 << " "<< r2 << " "<< r3 << std::endl;
  
  // Create a simple jet with 3 particles out of template momenta
  fastjet::PseudoJet axis3b = fastjet::join(my3bTemplate);
  // Compute the actual jet planar flow and the template planar flow
//  cout << "Jet Pf = " << pf2(HadTopJet) << std::endl;
//  cout << " Template Pf = " << pf(axis3b) << std::endl; 
  }
  

  
   if (tagged2 == 0)
  {
    std::cout << "No 3-body substructure found in Leptonic Top" << std::endl;
  }  
  else
  {
  
   /// We extract substructure information from filtered jet  
  // std::cout << "The number of pieces is " << tagged2.pieces().size() << std::endl;

            
  const fastjet::TemplateTagger::StructureType & templ2 = tagged2.structure_of<fastjet::TemplateTagger>(); 
  
    std::stringstream ss; 
  ss << "(Ov3 = "<<tagged2.structure_of<fastjet::TemplateTagger>().ov() <<")";
  
     // for event display
  set_header_for_display2(gout, tagged2.eta(), tagged2.phi_std(), coneRadius, ss.str() );
 //   set_header_for_display(fout,0, 0, 6.28, ss.str() );

  std::vector<fastjet::PseudoJet> constituents = LepTopJet.constituents();
  output_hadron_to_display(gout,constituents);
 
 
// Extract the peak 3-body template
  std::vector<fastjet::PseudoJet> my3bTemplate = templ2.maxTempl();
  
   std::vector<double> radii = templateRadii(my3bTemplate, subConeRadius3);
 
 double r1 = radii[0], r2 = 15./my3bTemplate[1].perp() , r3 = 0.1;
 if (r2 < 0.01) r2 =0.01;
 
  circle(gout,tagged2,coneRadius,6, "title \"hTop\"");
  circle(gout,my3bTemplate[0],r1,3, "title \"3b-templ\"");
  circle(gout,my3bTemplate[1],r2,3, "notitle");
  circle(gout,my3bTemplate[2],r3,3, "notitle");
  
  
  fastjet::PseudoJet parent1 = tagged2.pieces()[0];
  fastjet::PseudoJet parent2 = tagged2.pieces()[1];
  fastjet::PseudoJet parent3 = tagged2.pieces()[2];

  std::cout << "Found suitable filtered subjets: " << std::endl;
  std::cout << " " << parent1 << std::endl;
  std::cout << " " << parent2 << std::endl;
  std::cout << " " << parent3 << std::endl;

  std::cout << "Total = " << std::endl;
  std::cout << " " << tagged2 << std::endl;
  std::cout << "(Ov3 = " << templ2.ov() <<")" << std::endl << std::endl;
  std::cout << " The best-matched (3-body) template is: "<< std::endl;     
  PrintJets(templ2.maxTempl());
 
  
  
  
  }
  
    set_footer_for_display(fout);
    
    set_footer_for_display2(gout);
  
  FILE *pipe = popen("gnuplot -persist sample_event_display.dat", "w");//open up a pipe to gnuplot
  pclose(pipe); //close gnuplot pipe
 FILE *pipe2 = popen("gnuplot -persist sample_event_display2.dat", "w");//open up a pipe to gnuplot
  pclose(pipe2); //close gnuplot pipe
 // Show me how long it took
  std::cout << "Time Elapsed: " << ((double)clock() - start) / CLOCKS_PER_SEC << std::endl;

  std::string mystr;
  std::cout << "Please enter (y) to continue or (q) to abort the program:  ";
  std::getline (cin, mystr);
  if (mystr == "q") return 0;
  
  }
  
  // a bit of memory cleaning
  for (unsigned int i=0; i < HadTopTaggers.size(); i++) delete HadTopTaggers[i];
  for (unsigned int i=0; i < LepTopTaggers.size(); i++) delete LepTopTaggers[i];
  return 0;
}


//----------------------------------------------------------------------
// does the actual work for printing out a jet
//----------------------------------------------------------------------
ostream & operator<<(ostream & ostr, const fastjet::PseudoJet & jet) {
  ostr << "pt, y, phi =" 
       << " " << setw(10) << jet.perp() 
       << " " << setw(6) <<  jet.rap()  
       << " " << setw(6) <<  jet.phi()  
       << ", mass = " << setw(10) << jet.m()
       << ", btag = " << jet.user_index();
  return ostr;
	
	
}
